--    This file is part of Alp gen.
--    Copyright 2021 Thomas "V10lator" Rohloff <v10lator@myway.de
--
--    Alp gen is free software: you can redistribute it and/or modify
--    it under the terms of the GNU General Public License as published by
--    the Free Software Foundation, either version 3 of the License, or
--    (at your option) any later version.
--
--    Alp gen is distributed in the hope that it will be useful,
--    but WITHOUT ANY WARRANTY; without even the implied warranty of
--    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--    GNU General Public License for more details.
--
--    You should have received a copy of the GNU General Public License
--    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

local mapgenutil = require "V10lator/mapgenutil"
local treegen = require "V10lator/treegen"
local layersutil = require "terrain/layersutil"
local maputil = require "maputil"
local pipeUtil = require "V10lator/pipeUtil"

function data() 

return {
	climate = "temperate.clima.lua",
	order = 0,
	name = _("Alps"),
	params = {
		{
			key = "scale",
			name = _("Scale"),
			values = { "1:4", "1:2", "1:1" },
			defaultIndex = 2,
			uiType = "SLIDER",
		},
		{
			key = "water",
			name = _("Water"),
			values = { "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "" },
			defaultIndex = 6,
			uiType = "SLIDER",
		},
		{
			key = "lakes",
			name = _("Lakes"),
			values = { "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "" },
			defaultIndex = 30,
			uiType = "SLIDER",
		},
		{
			key = "lakeSize",
			name = _("Lake size"),
			values = { "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "" },
			defaultIndex = 30,
			uiType = "SLIDER",
		},
		{
			key = "forests",
			name = _("Forest"),
			values = { "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "" },
			defaultIndex = 1,
			uiType = "SLIDER",
		},
		{
			key = "forestSize",
			name = _("Size of Forest"),
			values = { "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "" },
			defaultIndex = 7,
			uiType = "SLIDER",
		},
		{
			key = "hilliness",
			name = _("Hilliness"),
			values = { "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "" },
			defaultIndex = 4,
			uiType = "SLIDER",
		},
	},
	updateFn = function(params)
		-- local temperateassetsgen = dofile("./res/scripts/terrain/temperateassetsgen.lua")
		-- local mapgenutil = dofile("./res/scripts/terrain/mapgenutil.lua")
		-- local layersutil = dofile("./res/scripts/terrain/layersutil.lua")
		local result = {
			parallelFactor = 32,
			heightmapLayer = "HM",
			layers = layersutil.Layer.new(),
		}

		local sameSeed = math.random(1, 100000000)
		math.randomseed(sameSeed) -- reset seed to match desert generator
		
		-- #################
		-- #### CONFIG
		local scale = params.scale + 1
		do
			local table = {
				0.25,
				0.5,
				1
			}
			scale = table[scale]
		end
		local water = params.water / 30
		local noWater = water == 0
		local lakes = params.lakes / 30
		local lakeSize = params.lakeSize / 30
		local hilliness = params.hilliness / 30
		
		do
			pipeUtil.init()
			local pipe = pipeUtil.open("alps_scale", "SEND")
			pipeUtil.write(pipe, scale)
			pipeUtil.close(pipe)
			pipeUtil.cleanup()
		end
		
		game.config.settings.v10lator = { alps = { scale = scale } }
		
		
		local riverConfig = {
			depthScale = 1.5,
			maxOrder = 2,
			segmentLength = 2400,
			bounds = params.bounds,
			baseProbability = water * water * 2,
			minDist = water > 0.5 and 2 or 3,
		}
		
		local rivers = {}
		if not noWater then
			local start = mapgenutil.FindGoodRiverStart(params.bounds)
			mapgenutil.MakeRivers(rivers, riverConfig, 120000, start.pos, start.angle)
			
			local lakeConfig = {
				getLakePropability = function()
					return lakes + 0.2
				end,
				lakeSize = lakeSize * 600, 
			}
			mapgenutil.MakeLakesOld(rivers, lakeConfig)
			
			-- local curvesConfig = {
				-- getStrength = function(position) 
					-- return 0.7
				-- end,
				-- getWidthMultiplier = function(position) 
					-- return 1
				-- end
			-- }
			-- mapgenutil.MakeCurvesOld(rivers, curvesConfig)
			
			maputil.Convert(rivers)
			maputil.ValidateRiver(rivers)
			-- maputil.PrintRiver(rivers) 
		end
	
		local ridgesConfig = {
			bounds = params.bounds, 
			probability = hilliness,
			minHeight = 1502 * scale, 
			maxHeight = 4697 * scale,
		}

		local valleys = {}
		local ridges = mapgenutil.MakeRidges(ridgesConfig)
		
		-- #################
		-- #### PREPARE
		local mkTemp = layersutil.TempMaker.new()
		
		-- #################
		-- #### BACKGROUND AND RIVER
		result.layers:Constant(result.heightmapLayer, .01)
		
		do -- river
			result.layers:PushColor("#0022DD")
			result.layers:River(result.heightmapLayer, rivers)
			result.layers:PopColor()
		end
		
		local noiseMap = mkTemp:Get()
		result.layers:Noise(noiseMap, 100)
		
		local distanceMap = mkTemp:Get()
		result.layers:Distance(result.heightmapLayer, distanceMap)
		
		-- #################
		-- #### RIDGES
		local ridgesMap
		do -- ridges
			result.layers:PushColor("#22AADD")
			local t1 = mkTemp:Get()
			result.layers:Map(distanceMap, t1, {15, 2500}, {0, 1}, true)
		--	result.layers:Mad(t1, noiseMap, result.heightmapLayer)
		--	noiseMap = mkTemp:Restore(noiseMap)
			
			local t2 = mkTemp:Get()
			result.layers:Ridge(t2, {
				valleys = valleys.points,
				ridges = ridges,
				noiseStrength = 0
			})
			
			result.layers:Map(distanceMap, t1, {50, 1500}, {0, 1}, true)
			ridgesMap = mkTemp:Get()
			result.layers:Mul(t1, t2, ridgesMap)
			t1 = mkTemp:Restore(t1)
			t2 = mkTemp:Restore(t2)

			result.layers:Add(ridgesMap, result.heightmapLayer, result.heightmapLayer)
			result.layers:PopColor()
		end
		
		-- #################
		-- #### NOISE
		local noiseStrength = 23 * scale
		result.layers:PushColor("#5577DD")
		local t2 = mkTemp:Get()
		result.layers:RidgedNoise(t2, { octaves = 5, frequency = 1 / 444, lacunarity = 2.2, gain = 0.8 } )
		result.layers:Map(t2, t2, {0, 4}, {0, noiseStrength * 1.2}, false)
		
		local t1 = mkTemp:Get()
		result.layers:Map(distanceMap, t1, {0, 30}, {0, 1}, true)
		
		result.layers:Mad(t2, t1, result.heightmapLayer)
		mkTemp:Restore(t1)
		mkTemp:Restore(t2)
		result.layers:PopColor()
		
		-- #################
		-- #### ASSETS
		result.layers:PushColor("#007777")
		result.forestMap, result.treesMapping, result.assetsMap, result.assetsMapping = treegen.Make(
			result.layers, mkTemp, result.heightmapLayer, ridgesMap, distanceMap, scale, noWater, params.forests / 30, params.forestSize / 30
		)
		result.layers:PopColor()
		distanceMap = nil
	
		-- #################
		-- #### FINISH
		ridgesMap = mkTemp:Restore(ridgesMap)
		mkTemp:Restore(result.forestMap)
		mkTemp:Restore(result.assetsMap)
		mkTemp:Finish()
		-- maputil.PrintGraph(result)
	
		return result
	end
}

end
